#include <iostream>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>
#include <GL/glx.h>

// this may possibly limit portability, some fallback is in order
#include <X11/extensions/xf86vmode.h>

#include "init.hpp"
#include "error_msg.h"

Display *dpy;
static Window win;
static GLXContext glx_ctx;
static XF86VidModeModeInfo orig_mode;
static bool fullscreen;

int InitGraphics(InitParams *iparams) {
	
	// open connection to the X server, 0 means use DISPLAY env var.
	if(!(dpy = XOpenDisplay(0))) {
		error("XOpenDisplay() failed");
		return -1;
	}

	int screen = DefaultScreen(dpy);
	Window root_win = RootWindow(dpy, screen);

	int color_bits = iparams->bpp == 15 || iparams->bpp == 16 ? 5 : 8;
	int stencil_bits = color_bits == 8 ? 8 : 1;
		
	if(iparams->zbuffer == 32) {
		iparams->zbuffer = 24;
	}

	int glx_attrib[] = {
		GLX_RGBA, GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, color_bits,
		GLX_GREEN_SIZE, color_bits,
		GLX_BLUE_SIZE, color_bits,
		GLX_DEPTH_SIZE, iparams->zbuffer,
		GLX_STENCIL_SIZE, stencil_bits,
		None
	};

	info("Initializing graphics system");
	info("Requested video mode: %dx%d %d bpp, %d zbuffer, %d stencil", iparams->x, iparams->y, iparams->bpp, iparams->zbuffer, stencil_bits);
	//Info("Aquired video mode: %dx%d %d bpp, %d zbuffer, %d stencil", iparams->x, iparams->y, iparams->bpp, iparams->zbuffer, stencil);
	
	XVisualInfo *vis_info;
	if(!(vis_info = glXChooseVisual(dpy, screen, glx_attrib))) {
		error("glXChooseVisual() failed");
		XCloseDisplay(dpy);
		return -1;
	}

	if(!(glx_ctx = glXCreateContext(dpy, vis_info, 0, True))) {
		error("glXCreateContext() failed");
		XFree(vis_info);
		XCloseDisplay(dpy);
		return -1;
	}


	XSetWindowAttributes xattr;
	xattr.colormap = XCreateColormap(dpy, root_win, vis_info->visual, AllocNone);

	if(iparams->fullscreen) {
		XF86VidModeModeInfo **modes;
		XF86VidModeModeInfo *vid_mode = 0;
		int mode_count;
		XF86VidModeGetAllModeLines(dpy, screen, &mode_count, &modes);
		orig_mode = *modes[0];

		for(int i=0; i<mode_count; i++) {
			if(modes[i]->hdisplay == iparams->x && modes[i]->vdisplay == iparams->y) {
				vid_mode = modes[i];
			}
		}
		if(!vid_mode) {
			error("Could not set requested video mode");
			XFree(modes);
			XFree(vis_info);
			XCloseDisplay(dpy);
			return -1;
		}
		
		XF86VidModeSwitchToMode(dpy, screen, vid_mode);
		XF86VidModeSetViewPort(dpy, screen, 0, 0);
		XFree(modes);

		xattr.override_redirect = True;
		win = XCreateWindow(dpy, root_win, 0, 0, iparams->x, iparams->y, 0, vis_info->depth, InputOutput, vis_info->visual, CWColormap | CWOverrideRedirect, &xattr);
		XSelectInput(dpy, win, ExposureMask | StructureNotifyMask | KeyPressMask | KeyReleaseMask);

		XWarpPointer(dpy, None, win, 0, 0, 0, 0, 0, 0);
		XMapRaised(dpy, win);
        XGrabKeyboard(dpy, win, True, GrabModeAsync, GrabModeAsync, CurrentTime);
        XGrabPointer(dpy, win, True, ButtonPressMask, GrabModeAsync, GrabModeAsync, win, None, CurrentTime);
		
		fullscreen = true;
	} else {
		win = XCreateWindow(dpy, root_win, 0, 0, iparams->x, iparams->y, 0, vis_info->depth, InputOutput, vis_info->visual, CWColormap, &xattr);
		XSelectInput(dpy, win, ExposureMask | StructureNotifyMask | KeyPressMask | KeyReleaseMask);
	}
		

	/* set WM cooperation settings */
	Atom wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", True);
	XSetWMProtocols(dpy, win, &wm_delete, 1);

	XTextProperty tp_wname;
	static char *window_title = "Sense Amok Graphics System on X";
	XStringListToTextProperty(&window_title, 1, &tp_wname);
	XSetWMName(dpy, win, &tp_wname);
	XFree(tp_wname.value);

	XClassHint class_hint;
	class_hint.res_name = "sense_amok";
	class_hint.res_class = "sende_amok_gfx";
	XSetClassHint(dpy, win, &class_hint);
	
	XFree(vis_info);

	if(glXMakeCurrent(dpy, win, glx_ctx) == False) {
		error("glXMakeCurrent() failed");
		glXDestroyContext(dpy, glx_ctx);
		XDestroyWindow(dpy, win);
		XCloseDisplay(dpy);
		return -1;
	}

	if(!glXIsDirect(dpy, glx_ctx)) {
		warning("using indirect rendering! (slow)");
	}

	XMapWindow(dpy, win);
	XFlush(dpy);
	
	return 0;
}


void DestroyGraphics() {
	glXMakeCurrent(dpy, None, 0);
	glXDestroyContext(dpy, glx_ctx);
	XDestroyWindow(dpy, win);

	if(fullscreen) {
		XF86VidModeSwitchToMode(dpy, DefaultScreen(dpy), &orig_mode);
		XF86VidModeSetViewPort(dpy, DefaultScreen(dpy), 0, 0);
	}
	
	XCloseDisplay(dpy);
}


void Flip() {
	glXSwapBuffers(dpy, win);
}
